from typing import Generic, TypeVar

from rainbowadn.flow.core import *
from rainbowadn.flow.sequence import *
from rainbowadn.nullability import *

__all__ = ('StackedDispatch',)

Element = TypeVar('Element')
Out = TypeVar('Out')


class StackedDispatch(SequenceDispatch[Element, Out], Generic[Element, Out]):
    def __init__(self, stacked: Reduce[tuple[Nullable[Element], Element], Out], out: Out):
        assert isinstance(stacked, Reduce)
        self.stacked = stacked
        self.out = out

    async def on_first(self, element: Element) -> Out:
        return await self.stacked.reduce(self.out, (Null(), element))

    async def on_last(self, element: Element) -> Out:
        return self.out

    async def on_pair(self, previous: Element, element: Element) -> Out:
        return await self.stacked.reduce(self.out, (NotNull(previous), element))
