from typing import Generic, TypeVar

from rainbowadn.flow.core import *
from rainbowadn.flow.sequence import *
from rainbowadn.nullability import *
from ._stackedreduce import *

__all__ = ('StackedReducer',)

Stacked = TypeVar('Stacked')
Out = TypeVar('Out')


class StackedReducer(
    Reducer[tuple[Nullable[Stacked], Stacked], Out],
    Generic[Out, Stacked]
):
    def __init__(self, stacked: Reducer[SequenceDispatcher[Stacked, Out], Out]):
        assert isinstance(stacked, Reducer)
        self.stacked = stacked

    async def reduce(self, reduce: Reduce[tuple[Nullable[Stacked], Stacked], Out]) -> Out:
        assert isinstance(reduce, Reduce)
        stacked_reduce: StackedReduce[Out, Stacked] = StackedReduce(reduce)
        dispatcher_reduce: Reduce[SequenceDispatcher[Stacked, Out], Out] = stacked_reduce.loose()
        assert isinstance(dispatcher_reduce, Reduce)
        return await self.stacked.reduce(dispatcher_reduce)

    def loose(self) -> Reducer[tuple[Nullable[Stacked], Stacked], Out]:
        return self

    @classmethod
    def mapper(cls) -> Mapper[
        Reducer[SequenceDispatcher[Stacked, Out], Out],
        Reducer[tuple[Nullable[Stacked], Stacked], Out],
    ]:
        return CallableMapper(cls)
