from typing import Generic, TypeVar

from rainbowadn.core import *
from rainbowadn.flow.core import *
from rainbowadn.flow.sequence import *
from rainbowadn.flow.verification.core import *
from rainbowadn.nullability import *
from ._stageverificationprotocol import *

__all__ = ('StageVerificationDispatch',)

Base = TypeVar('Base')
Stage = TypeVar('Stage')
Header = TypeVar('Header')


class StageVerificationDispatch(
    SequenceDispatch[Stage, bool],
    Generic[Base, Stage, Header]
):
    def __init__(
            self,
            protocol: StageVerificationProtocol[Base, Stage, Header],
            previous: Nullable[Base],
            header: Header,
            base: Base,
    ):
        assert isinstance(protocol, StageVerificationProtocol)
        assert isinstance(previous, Nullable)
        self.protocol = protocol
        self.previous = previous
        self.header = header
        self.base = base

    async def on_first(self, element: Stage) -> bool:
        assert_true(await self.protocol.initial(self.previous, self.header, element))
        return True

    async def on_last(self, element: Stage) -> bool:
        assert_true(await self.protocol.final(element, self.base))
        return True

    async def on_pair(self, previous: Stage, element: Stage) -> bool:
        assert_true(await self.protocol.intermediate(previous, element))
        return True

    def _verification(self) -> Verification[SequenceDispatcher[Stage, bool]]:
        return MapperVerification(DispatchMapper(self).loose())

    def reduce_verification(self) -> Verification[Reducer[SequenceDispatcher[Stage, bool], bool]]:
        return ReduceVerification(self._verification()).loose()
